(function (global, factory) {
    typeof exports === 'object' && typeof module !== 'undefined' ? factory(exports) :
    typeof define === 'function' && define.amd ? define(['exports'], factory) :
    (global = typeof globalThis !== 'undefined' ? globalThis : global || self, factory(global.vella = {}));
})(this, (function (exports) { 'use strict';

    function getErrorMessage(id) {
        return `[${id}]: ${messages[id].message}`;
    }
    const messages = {
        "A001": {
            message: "can't call this hook from here",
            explanation: `
Lifecycle hooks in vella are static functions.
The engine will prevent them from being registered in contexts that are not lifecycle related.
In lifecycle context, hooks can't be registered later than the time at which they can fire.
For example:

- \`removing\` and \`reflowed\` can be called from \`asap\`, but the converse isn't true
- likewise, you can call \`removing\` from \`reflowed\`, but you can't call them th other way around.
`,
            versionRange: { from: "v0.0.5" }
        },
        "A002": {
            message: "boot expects an element or a POJO specifying the nextSibling as first parameter",
            explantaion: "TODO",
            versionRange: { from: "v0.0.5" }
        },
        "A003": {
            message: "boot expects a live zone as second parameter",
            explantaion: "TODO",
            versionRange: { from: "v0.0.5" }
        },
        "A004": {
            message: "the removeManager can only be set once per liveZone",
            explanation: "",
            versionRange: { from: "v0.0.5" }
        },
        "A005": {
            message: "keyed expects a stream or a live zone of keys as first argument",
            explanation: "",
            versionRange: { from: "v0.0.5" }
        },
        "A006": {
            message: "keyed expects a renderer as second argument",
            explanation: "",
            versionRange: { from: "v0.0.5" }
        },
    };

    let doc;
    let win;
    let tagCache;
    function setWindow(w) {
        doc = w.document;
        w.Node.prototype;
        tagCache = { "": {}, "http://www.w3.org/2000/svg": {}, "http://www.w3.org/1998/Math/MathML": {} };
        win = w;
    }
    /* c8 ignore next 3*/
    if (typeof window != "undefined")
        setWindow(window);
    // eslint-disable-next-line @typescript-eslint/no-explicit-any, @typescript-eslint/no-non-null-assertion
    win != null && !!win.chrome;

    // Public interface
    var S$1 = function S(fn, value) {
        if (Owner === null)
            console.warn("computations created without a root or parent will never be disposed");
        var { node, value: _value } = makeComputationNode(fn, value, false, false);
        if (node === null) {
            return function computation() { return _value; };
        }
        else {
            return function computation() {
                return node.current();
            };
        }
    };
    // compatibility with commonjs systems that expect default export to be at require('s.js').default rather than just require('s-js')
    Object.defineProperty(S$1, 'default', { value: S$1 });
    const warn = { conflict: true };
    S$1.dontWarn = () => { warn.conflict = false; };
    S$1.root = function root(fn) {
        var owner = Owner, disposer = fn.length === 0 ? null : function _dispose() {
            if (root === null) ;
            else if (RunningClock !== null) {
                RootClock.disposes.add(root);
            }
            else {
                dispose(root);
            }
        }, root = disposer === null ? UNOWNED : getCandidateNode(), result;
        Owner = root;
        try {
            result = disposer === null ? fn() : fn(disposer);
        }
        finally {
            Owner = owner;
        }
        if (disposer !== null && recycleOrClaimNode(root, null, undefined, true)) {
            root = null;
        }
        return result;
    };
    S$1.on = function on(ev, fn, seed, onchanges) {
        if (Array.isArray(ev))
            ev = callAll(ev);
        onchanges = !!onchanges;
        return S$1(on, seed);
        function on(value) {
            var listener = Listener;
            ev();
            if (onchanges)
                onchanges = false;
            else {
                Listener = null;
                value = fn(value);
                Listener = listener;
            }
            return value;
        }
    };
    function callAll(ss) {
        return function all() {
            for (var i = 0; i < ss.length; i++)
                ss[i]();
        };
    }
    S$1.effect = function effect(fn, value) {
        makeComputationNode(fn, value, false, false);
    };
    S$1.data = function data(value) {
        var node = new DataNode(value);
        return function data(value) {
            if (arguments.length === 0) {
                return node.current();
            }
            else {
                return node.next(value);
            }
        };
    };
    S$1.value = function value(current, eq) {
        var node = new DataNode(current), age = -1;
        return function value(update) {
            if (arguments.length === 0) {
                return node.current();
            }
            else {
                var same = eq ? eq(current, update) : current === update;
                if (!same) {
                    var time = RootClock.time;
                    if (age === time && warn.conflict) {
                        console.warn("conflicting values: " + update + " is not the same as " + current);
                    }
                    age = time;
                    current = update;
                    node.next(update);
                }
                return update;
            }
        };
    };
    S$1.freeze = function freeze(fn) {
        var result = undefined;
        if (RunningClock !== null) {
            result = fn();
        }
        else {
            RunningClock = RootClock;
            RunningClock.changes.reset();
            try {
                result = fn();
                event();
            }
            finally {
                RunningClock = null;
            }
        }
        return result;
    };
    S$1.sample = function sample(fn) {
        var result, listener = Listener;
        Listener = null;
        result = fn();
        Listener = listener;
        return result;
    };
    S$1.cleanup = function cleanup(fn) {
        if (Owner === null)
            console.warn("cleanups created without a root or parent will never be run");
        else if (Owner.cleanups === null)
            Owner.cleanups = [fn];
        else
            Owner.cleanups.push(fn);
    };
    // experimental : exposing node constructors and some state
    S$1.makeDataNode = function makeDataNode(value) {
        return new DataNode(value);
    };
    S$1.makeComputationNode = makeComputationNode;
    S$1.disposeNode = function disposeNode(node) {
        if (RunningClock !== null) {
            RootClock.disposes.add(node);
        }
        else {
            dispose(node);
        }
    };
    S$1.isFrozen = function isFrozen() {
        return RunningClock !== null;
    };
    S$1.isListening = function isListening() {
        return Listener !== null;
    };
    // Internal implementation
    /// Graph classes and operations
    class Clock {
        constructor() {
            this.time = 0;
            this.changes = new Queue(); // batched changes to data nodes
            this.updates = new Queue(); // computations to update
            this.disposes = new Queue(); // disposals to run after current batch of updates finishes
        }
    }
    var RootClockProxy = {
        time: function () { return RootClock.time; }
    };
    class DataNode {
        constructor(value) {
            this.value = value;
            this.pending = NOTPENDING;
            this.log = null;
        }
        current() {
            if (Listener !== null) {
                logDataRead(this);
            }
            return this.value;
        }
        next(value) {
            if (RunningClock !== null) {
                if (this.pending !== NOTPENDING) { // value has already been set once, check for conflicts
                    if (value !== this.pending && warn.conflict) {
                        console.warn("conflicting changes: " + value + " !== " + this.pending);
                    }
                }
                else { // add to list of changes
                    this.pending = value;
                    RootClock.changes.add(this);
                }
            }
            else { // not batching, respond to change now
                if (this.log !== null) {
                    this.pending = value;
                    RootClock.changes.add(this);
                    event();
                }
                else {
                    this.value = value;
                }
            }
            return value;
        }
        clock() {
            return RootClockProxy;
        }
    }
    class ComputationNode {
        constructor() {
            this.fn = null;
            this.value = undefined;
            this.age = -1;
            this.state = CURRENT;
            this.source1 = null;
            this.source1slot = 0;
            this.sources = null;
            this.sourceslots = null;
            this.log = null;
            this.owned = null;
            this.cleanups = null;
        }
        current() {
            if (Listener !== null) {
                if (this.age === RootClock.time) {
                    if (this.state === RUNNING)
                        throw new Error("circular dependency");
                    else
                        updateNode(this); // checks for state === STALE internally, so don't need to check here
                }
                logComputationRead(this);
            }
            return this.value;
        }
        clock() {
            return RootClockProxy;
        }
    }
    class Log {
        constructor() {
            this.node1 = null;
            this.node1slot = 0;
            this.nodes = null;
            this.nodeslots = null;
        }
    }
    class Queue {
        constructor() {
            this.items = [];
            this.count = 0;
        }
        reset() {
            this.count = 0;
        }
        add(item) {
            this.items[this.count++] = item;
        }
        run(fn) {
            var items = this.items;
            for (var i = 0; i < this.count; i++) {
                fn(items[i]);
                items[i] = null;
            }
            this.count = 0;
        }
    }
    // Constants
    var NOTPENDING = {}, CURRENT = 0, STALE = 1, RUNNING = 2, UNOWNED = new ComputationNode();
    // "Globals" used to keep track of current system state
    var RootClock = new Clock(), RunningClock = null, // currently running clock 
    Listener = null, // currently listening computation
    Owner = null, // owner for new computations
    LastNode = null; // cached unused node, for re-use
    // Functions
    var makeComputationNodeResult = { node: null, value: undefined };
    function makeComputationNode(fn, value, orphan, sample) {
        var node = getCandidateNode(), owner = Owner, listener = Listener, toplevel = RunningClock === null;
        Owner = node;
        Listener = sample ? null : node;
        if (toplevel) {
            value = execToplevelComputation(fn, value);
        }
        else {
            value = fn(value);
        }
        Owner = owner;
        Listener = listener;
        var recycled = recycleOrClaimNode(node, fn, value, orphan);
        if (toplevel)
            finishToplevelComputation(owner, listener);
        makeComputationNodeResult.node = recycled ? null : node;
        makeComputationNodeResult.value = value;
        return makeComputationNodeResult;
    }
    function execToplevelComputation(fn, value) {
        RunningClock = RootClock;
        RootClock.changes.reset();
        RootClock.updates.reset();
        try {
            return fn(value);
        }
        finally {
            Owner = Listener = RunningClock = null;
        }
    }
    function finishToplevelComputation(owner, listener) {
        if (RootClock.changes.count > 0 || RootClock.updates.count > 0) {
            RootClock.time++;
            try {
                run(RootClock);
            }
            finally {
                RunningClock = null;
                Owner = owner;
                Listener = listener;
            }
        }
    }
    function getCandidateNode() {
        var node = LastNode;
        if (node === null)
            node = new ComputationNode();
        else
            LastNode = null;
        return node;
    }
    function recycleOrClaimNode(node, fn, value, orphan) {
        var _owner = orphan || Owner === null || Owner === UNOWNED ? null : Owner, recycle = node.source1 === null && (node.owned === null && node.cleanups === null || _owner !== null), i;
        if (recycle) {
            LastNode = node;
            if (_owner !== null) {
                if (node.owned !== null) {
                    if (_owner.owned === null)
                        _owner.owned = node.owned;
                    else
                        for (i = 0; i < node.owned.length; i++) {
                            _owner.owned.push(node.owned[i]);
                        }
                    node.owned = null;
                }
                if (node.cleanups !== null) {
                    if (_owner.cleanups === null)
                        _owner.cleanups = node.cleanups;
                    else
                        for (i = 0; i < node.cleanups.length; i++) {
                            _owner.cleanups.push(node.cleanups[i]);
                        }
                    node.cleanups = null;
                }
            }
        }
        else {
            node.fn = fn;
            node.value = value;
            node.age = RootClock.time;
            if (_owner !== null) {
                if (_owner.owned === null)
                    _owner.owned = [node];
                else
                    _owner.owned.push(node);
            }
        }
        return recycle;
    }
    function logRead(from) {
        var to = Listener, fromslot, toslot = to.source1 === null ? -1 : to.sources === null ? 0 : to.sources.length;
        if (from.node1 === null) {
            from.node1 = to;
            from.node1slot = toslot;
            fromslot = -1;
        }
        else if (from.nodes === null) {
            from.nodes = [to];
            from.nodeslots = [toslot];
            fromslot = 0;
        }
        else {
            fromslot = from.nodes.length;
            from.nodes.push(to);
            from.nodeslots.push(toslot);
        }
        if (to.source1 === null) {
            to.source1 = from;
            to.source1slot = fromslot;
        }
        else if (to.sources === null) {
            to.sources = [from];
            to.sourceslots = [fromslot];
        }
        else {
            to.sources.push(from);
            to.sourceslots.push(fromslot);
        }
    }
    function logDataRead(data) {
        if (data.log === null)
            data.log = new Log();
        logRead(data.log);
    }
    function logComputationRead(node) {
        if (node.log === null)
            node.log = new Log();
        logRead(node.log);
    }
    function event() {
        // b/c we might be under a top level S.root(), have to preserve current root
        var owner = Owner;
        RootClock.updates.reset();
        RootClock.time++;
        try {
            run(RootClock);
        }
        finally {
            RunningClock = Listener = null;
            Owner = owner;
        }
    }
    function run(clock) {
        var running = RunningClock, count = 0;
        RunningClock = clock;
        clock.disposes.reset();
        // for each batch ...
        while (clock.changes.count !== 0 || clock.updates.count !== 0 || clock.disposes.count !== 0) {
            if (count > 0) // don't tick on first run, or else we expire already scheduled updates
                clock.time++;
            clock.changes.run(applyDataChange);
            clock.updates.run(updateNode);
            clock.disposes.run(dispose);
            // if there are still changes after excessive batches, assume runaway            
            if (count++ > 1e5) {
                throw new Error("Runaway clock detected");
            }
        }
        RunningClock = running;
    }
    function applyDataChange(data) {
        data.value = data.pending;
        data.pending = NOTPENDING;
        if (data.log)
            markComputationsStale(data.log);
    }
    function markComputationsStale(log) {
        var node1 = log.node1, nodes = log.nodes;
        // mark all downstream nodes stale which haven't been already
        if (node1 !== null)
            markNodeStale(node1);
        if (nodes !== null) {
            for (var i = 0, len = nodes.length; i < len; i++) {
                markNodeStale(nodes[i]);
            }
        }
    }
    function markNodeStale(node) {
        var time = RootClock.time;
        if (node.age < time) {
            node.age = time;
            node.state = STALE;
            RootClock.updates.add(node);
            if (node.owned !== null)
                markOwnedNodesForDisposal(node.owned);
            if (node.log !== null)
                markComputationsStale(node.log);
        }
    }
    function markOwnedNodesForDisposal(owned) {
        for (var i = 0; i < owned.length; i++) {
            var child = owned[i];
            child.age = RootClock.time;
            child.state = CURRENT;
            if (child.owned !== null)
                markOwnedNodesForDisposal(child.owned);
        }
    }
    function updateNode(node) {
        if (node.state === STALE) {
            var owner = Owner, listener = Listener;
            Owner = Listener = node;
            node.state = RUNNING;
            cleanup(node, false);
            node.value = node.fn(node.value);
            node.state = CURRENT;
            Owner = owner;
            Listener = listener;
        }
    }
    function cleanup(node, final) {
        var source1 = node.source1, sources = node.sources, sourceslots = node.sourceslots, cleanups = node.cleanups, owned = node.owned, i, len;
        if (cleanups !== null) {
            for (i = 0; i < cleanups.length; i++) {
                cleanups[i](final);
            }
            node.cleanups = null;
        }
        if (owned !== null) {
            for (i = 0; i < owned.length; i++) {
                dispose(owned[i]);
            }
            node.owned = null;
        }
        if (source1 !== null) {
            cleanupSource(source1, node.source1slot);
            node.source1 = null;
        }
        if (sources !== null) {
            for (i = 0, len = sources.length; i < len; i++) {
                cleanupSource(sources.pop(), sourceslots.pop());
            }
        }
    }
    function cleanupSource(source, slot) {
        var nodes = source.nodes, nodeslots = source.nodeslots, last, lastslot;
        if (slot === -1) {
            source.node1 = null;
        }
        else {
            last = nodes.pop();
            lastslot = nodeslots.pop();
            if (slot !== nodes.length) {
                nodes[slot] = last;
                nodeslots[slot] = lastslot;
                if (lastslot === -1) {
                    last.source1slot = slot;
                }
                else {
                    last.sourceslots[lastslot] = slot;
                }
            }
        }
    }
    function dispose(node) {
        node.fn = null;
        node.log = null;
        cleanup(node, true);
    }

    function map(cb) {
        if (typeof cb !== "function")
            console.log(cb);
        return S_(() => cb(this()));
    }
    function toJSON() {
        return S$1.sample(this);
    }
    // eslint-disable-next-line @typescript-eslint/ban-types
    function decorateS(f) {
        return (...args) => {
            const res = f(...args);
            if (typeof res === "function") {
                res.map = map;
                res.toJSON = toJSON;
            }
            return res;
        };
    }
    const S_ = decorateS(S$1);
    Object.keys(S$1).forEach((m) => {
        S_[m] = (m === "cleanup" || m === "sample" || m === "root" || m === "freeze") ? S$1[m] : decorateS(S$1[m]);
    });

    // gets down to the value of nested streams.
    // in other words `flatten`, but where's the fun with that?
    const objProto = Object.prototype;
    const hasOwn = objProto.hasOwnProperty;
    const getProto = Object.getPrototypeOf;
    function skippable(x) {
        return x == null || x === true || x === false;
    }

    /* eslint-disable @typescript-eslint/ban-types */
    // Components
    // ==========
    const componentEarmark = Symbol("component earmark");
    function component(f, ...args) {
        const res = f.bind(null, ...args);
        res[componentEarmark] = true;
        return res;
    }
    let DOM;
    let Range;
    let Zone;
    let NodeCount = 0;
    let FirstInserted;
    let LastInserted;
    const setRange = (nr) => (Range = nr);
    const setZone = (zn) => (Zone = zn);
    function withoutRange(cb) {
        const range = Range;
        const firstInserted = FirstInserted;
        const nodeCount = NodeCount;
        Range = FirstInserted = null;
        try {
            cb();
        }
        finally {
            Range = range;
            FirstInserted = firstInserted;
            NodeCount = nodeCount;
        }
    }
    function withRange(nr, cb) {
        if (nr == null)
            return withoutRange(cb);
        const range = Range;
        const firstInserted = FirstInserted;
        const nodeCount = NodeCount;
        Range = nr;
        FirstInserted = null;
        NodeCount = 0;
        try {
            cb();
        }
        finally {
            nr.firstNode = FirstInserted;
            nr.lastNode = LastInserted;
            nr.nodeCount = NodeCount;
            if (firstInserted != null)
                FirstInserted = firstInserted;
            Range = range;
            NodeCount += nodeCount;
        }
    }
    function withRef(ref, cb) {
        const dom = DOM;
        DOM = ref;
        try {
            return cb();
        }
        finally {
            DOM = dom;
        }
    }
    // helpers
    function forEachNode(nr, cb, includingComments = false, result) {
        const needsMetadata = cb.length > 1;
        const lastFragmentIndex = nr.nodeCount - 1;
        if (lastFragmentIndex === -1)
            return;
        S_.freeze(() => {
            const { firstNode, lastNode } = nr;
            const boundary = lastNode.nextSibling;
            let node = firstNode;
            let fragmentIndex = 0;
            do {
                const next = node.nextSibling;
                if (includingComments || node.nodeType !== 8)
                    try {
                        const x = needsMetadata
                            ? cb(node, { fragmentIndex, lastFragmentIndex })
                            : cb(node);
                        if (result != null)
                            result.push(x);
                    }
                    finally { /**/ }
                fragmentIndex++;
                node = next;
            } while (node !== boundary);
        });
    }
    // constructors
    function DOMRef(parent, nextSibling) {
        return { parent, nextSibling };
    }
    function NodeRange({ parentNode, parentNodeRange } = {}) {
        if (parentNode == null)
            parentNode = DOM.parent;
        const res = {
            parentNode,
            parentNodeRange,
            firstNode: null,
            lastNode: null,
            nodeCount: 0,
            removeHooks: null
        };
        // Object.defineProperties(res, {
        // 	firstNode: {
        // 		set(x){
        // 			this._fN = x;
        // 			console.trace("SET fN", x && x.textContent)
        // 		},
        // 		get() {return this._fN}
        // 	},
        // 	lastNode: {
        // 		set(x){this._lN = x; console.trace("SET lN", x && x.textContent)},
        // 		get() {return this._lN}
        // 	},
        // })
        return res;
    }
    function fromOld({ parentNode, parentNodeRange }) {
        return NodeRange({ parentNode, parentNodeRange });
    }
    function fromParent(parentNodeRange) {
        const { parentNode } = parentNodeRange || {};
        return NodeRange({ parentNode, parentNodeRange: parentNodeRange });
    }
    function insert(node) {
        if (DOM.nextSibling != null)
            DOM.parent.insertBefore(node, DOM.nextSibling);
        else
            DOM.parent.appendChild(node);
        if (Range != null) {
            if (FirstInserted == null)
                FirstInserted = node;
            NodeCount++;
            LastInserted = node;
        }
    }
    function remove(nr) {
        const { parentNode } = nr;
        // TODO: revisit taking the placeholder comment into account
        if (nr.firstNode != nr.lastNode && parentNode.firstChild === nr.firstNode && parentNode.lastChild === nr.lastNode) {
            parentNode.textContent = "";
            return;
        }
        // TODO: check why node is implicitly "any"
        forEachNode(nr, (node) => {
            if (node.parentNode != null)
                parentNode.removeChild(node);
        }, true);
    }
    function syncParents(nr, key, old, current) {
        if (nr != null && nr[key] === old) {
            nr[key] = current;
            syncParents(nr.parentNodeRange, key, old, current);
        }
    }
    function emit(node) {
        if (skippable(node))
            return;
        if (Array.isArray(node))
            node.forEach(n => { emit(n); });
        else if (typeof node === "function") {
            if (hasOwn.call(node, componentEarmark))
                emit(node());
            else
                emitDynamic(node);
        }
        else {
            //TODO: workaround for making actual DOM fragments permanent
            if (isNode(node))
                insert(node);
            else
                insert(doc.createTextNode(String(node)));
        }
    }
    function isNode(node) {
        return !!node?.nodeType;
    }
    //stubs
    function emitDynamic(cb) {
        const lastNodes = new Set();
        let lastNr, placeHolderComment, remover;
        S_(() => {
            // Not sure this is needed given how S.js works.
            // It would be nice to let users set a recursion limit.
            const nr = setZone(lastNr == null ? fromParent(Range) : fromOld(lastNr));
            const nextSibling = lastNr != null ? lastNr.firstNode : DOM.nextSibling;
            const { parentNode } = nr;
            withRange(nr, () => withRef(DOMRef(parentNode, nextSibling), () => {
                try {
                    emit(cb());
                }
                finally {
                    const empty = FirstInserted == null;
                    const wasEmpty = lastNr != null && lastNr.firstNode === placeHolderComment;
                    // insert even if it was already present to keep the nodeRanges in sync
                    if (empty) {
                        if (placeHolderComment == null)
                            placeHolderComment = doc.createComment("");
                        insert(placeHolderComment);
                    }
                    // sync parent on redraw. LastNode is sinced when `remove` happens (either sync or async)
                    if (lastNr != null)
                        syncParents(nr.parentNodeRange, "firstNode", lastNr.firstNode, FirstInserted);
                    // Node removal can be racy if there are several concurrent `removing` phases that overlap, and
                    // if an earlier `removing` phase resolves after one that started later.
                    // When nodes are removed, It is important that a parent live zone is updated correctly if present.
                    // Since on update, the new nodes are inserted before the old ones, we can keep a stack of `lastNodes`
                    // on insertion, then, on removal, remove the one of the NodeRange that is being removed, and sync
                    // the parent with the oldest remaining `lastNode`.
                    lastNodes.add(LastInserted);
                    if (remover != null && !(empty && wasEmpty))
                        remover();
                    S_.cleanup(() => {
                        remover = () => {
                            if (nr.removeHooks != null) {
                                const results = [];
                                nr.removeHooks.hooks.forEach(x => x(results));
                                void (nr.removeHooks.manager || Promise.all.bind(Promise))(results).finally(() => {
                                    lastNodes.delete(nr.lastNode);
                                    const lastNode = lastNodes[Symbol.iterator]().next().value;
                                    remove(nr);
                                    syncParents(nr.parentNodeRange, "lastNode", nr.lastNode, lastNode);
                                });
                            }
                            else {
                                lastNodes.delete(nr.lastNode);
                                const lastNode = lastNodes[Symbol.iterator]().next().value;
                                remove(nr);
                                syncParents(nr.parentNodeRange, "lastNode", nr.lastNode, lastNode);
                            }
                        };
                    });
                    lastNr = nr;
                }
            }));
        });
    }
    function emitWithNodeRange(node) {
        const dr = fromParent(Range);
        withRange(dr, () => { emit(node); });
        return dr;
    }
    function boot(parentNode, main) {
        if (parentNode == null)
            throw new TypeError(getErrorMessage("A002"));
        let nextSibling;
        if (Object.getPrototypeOf(parentNode) === Object.prototype && parentNode.nextSibling != null) {
            nextSibling = parentNode.nextSibling;
            parentNode = nextSibling.parentNode;
        }
        if (String(parentNode.nodeType)[0] !== "1"
            || nextSibling != null && typeof nextSibling.nodeType !== "number")
            throw TypeError(getErrorMessage("A002"));
        if (typeof main !== "function" || hasOwn.call(main, componentEarmark))
            throw new TypeError(getErrorMessage("A003"));
        return S_.root(dispose => {
            const range = withRef(DOMRef(parentNode, nextSibling), () => emitWithNodeRange(main));
            return () => {
                dispose();
                remove(range);
            };
        });
    }

    /* eslint-disable @typescript-eslint/ban-types */
    let globalHooks;
    function Hooks() {
        return {
            beforeUpdating: null, afterUpdated: null, rendered: null, reflowed: null, updating: null, updated: null, removing: null
        };
    }
    const life = ["beforeUpdating", "afterUpdated", "rendered", "reflowed", "updating", "updated", "removing"].reduce((life, name) => {
        life[name] = function (cb) {
            if (globalHooks[name] == null)
                globalHooks[name] = [];
            // eslint-disable-next-line @typescript-eslint/no-non-null-assertion
            globalHooks[name].push(cb);
        };
        return life;
    }, {});
    function normalize(renderer) {
        if (renderer == null)
            throw new TypeError(getErrorMessage("A006"));
        if (typeof renderer === "function")
            return { render: renderer, hooks: null };
        return renderer;
    }
    function keyed(keys, hooks) {
        return component(Keyed, keys, normalize(hooks));
    }
    function Keyed(keys, { hooks, render }) {
        if (typeof hooks === "function") {
            globalHooks = Hooks();
            hooks({ ...life });
            hooks = globalHooks;
        }
        const parentDOMRange = Range;
        const { parent: parentNode } = DOM;
        const placeHolderComment = doc.createComment("");
        const hasIndices = render.length > 1;
        let final;
        S_(last => {
            final = false;
            S_.cleanup(() => {
                final = true;
                Promise.resolve().then(() => {
                    if (final && last != null)
                        last.refs.forEach((ref) => { ref.dispose(); });
                });
            });
            if (last === null) {
                return create(keys(), render, placeHolderComment, hasIndices);
            }
            else {
                const nextSibling = last.refs.length === 0
                    ? placeHolderComment
                    : last.refs[last.refs.length - 1].range.lastNode.nextSibling;
                // if (hooks != null && hooks.beforeUpdating != null) hooks.beforeUpdating.forEach(cb => {try {cb()}finally{/**/}})
                last = update(keys(), last, render, hooks, parentDOMRange, parentNode, nextSibling, placeHolderComment, hasIndices);
                // if (hooks != null && hooks.afterUpdated != null) hooks.afterUpdated.forEach(cb => {try {cb()}finally{/**/}})
                return last;
            }
        }, null);
    }
    const remover = (ref) => { remove(ref.range); ref.dispose(); };
    function moveNodes(parent, nextSibling, range) {
        if (range.firstNode === range.lastNode) {
            parent.insertBefore(range.firstNode, nextSibling);
        }
        else {
            const last = range.lastNode;
            let node = range.firstNode, next;
            // eslint-disable-next-line no-constant-condition
            while (true) {
                next = node.nextSibling;
                parent.insertBefore(node, nextSibling);
                if (next === last) {
                    parent.insertBefore(last, nextSibling);
                    break;
                }
                node = next;
            }
        }
    }
    // const first = x => x.range.firstNode.textContent
    function createMap(keys) {
        if (keys.length === 0)
            return null;
        const type = typeof keys[0];
        if (type === "string" || type === "number")
            return {
                kind: "object",
                map: keys.reduce((acc, k, i) => {
                    if (acc[k])
                        throw new Error(`duplicate key: ${k}`);
                    acc[k] = i;
                    return acc;
                }, Object.create(null))
            };
        else
            return {
                kind: "map",
                map: keys.reduce((acc, k, i) => {
                    if (acc.has(k)) {
                        console.error(k);
                        throw new Error("duplicate key (see above)");
                    }
                    acc.set(k, i);
                    return acc;
                }, new Map)
            };
    }
    function createListItem(refs, key, i, hasIndices, render) {
        return S_.root((dispose) => {
            const index = hasIndices ? S_.value(i) : null;
            refs[i] = { dispose, range: emitWithNodeRange(render(key, index)), index };
        });
    }
    function create(keys, render, placeHolderComment, hasIndices) {
        if (keys.length === 0) {
            insert(placeHolderComment);
            return { keys: [], refs: [] };
        }
        else {
            keys = [...keys];
            const refs = Array(keys.length);
            keys.forEach((key, i) => createListItem(refs, key, i, hasIndices, render));
            return { refs, keys };
        }
    }
    // Pilfered from Mithril v2
    function update(keys, 
    // data from the last render
    { keys: old, refs: oldRefs }, 
    //hooks
    render, hooks, 
    // DOM context
    parentNodeRange, parentNode, nextSibling, placeHolderComment, hasIndices) {
        keys = [...keys];
        const refs = Array(keys.length);
        if (hooks != null && hooks.updating != null) {
            const lastKeyIndex = old.length;
            old.forEach((key, i) => {
                hooks.updating.forEach(hook => {
                    const cb = hook.length > 1 ? (node, md) => hook(node, { ...md, key, keyIndex: i, lastKeyIndex }) : node => hook(node);
                    forEachNode(refs[i].range, cb);
                });
            });
        }
        if (old.length === 0) {
            if (keys.length > 0) {
                const range = Range;
                setRange(parentNodeRange);
                withRef(DOMRef(parentNode, nextSibling), () => {
                    keys.forEach((key, i) => createListItem(refs, key, i, hasIndices, render));
                });
                setRange(range);
                parentNode.removeChild(placeHolderComment);
                syncParents(parentNodeRange, "firstNode", placeHolderComment, refs[0].range.firstNode);
                syncParents(parentNodeRange, "lastNode", placeHolderComment, refs[refs.length - 1].range.lastNode);
            }
        }
        else if (keys.length === 0) {
            oldRefs.forEach((ref, i) => remover(ref)); //TODO: , old[i]))
            withRef(DOMRef(parentNode, nextSibling), () => {
                insert(placeHolderComment);
            });
            syncParents(parentNodeRange, "firstNode", oldRefs[0].range.firstNode, placeHolderComment);
            syncParents(parentNodeRange, "lastNode", oldRefs[oldRefs.length - 1].range.lastNode, placeHolderComment);
        }
        else {
            let os = 0, ks = 0, oe = old.length - 1, ke = keys.length - 1;
            // bottom-up
            while (oe >= os && ke >= ks && old[oe] === keys[ke]) {
                refs[ke] = oldRefs[oe];
                if (hasIndices)
                    refs[ke].index(ke);
                nextSibling = refs[ke].range.firstNode;
                oe--, ke--;
            }
            // top-down
            while (oe >= os && ke >= ks && old[os] === keys[ks]) {
                refs[ks] = oldRefs[os];
                if (hasIndices)
                    refs[ks].index(ks);
                os++, ks++;
            }
            // swaps and list reversals
            while (oe >= os && ke >= ks) {
                if (ks === ke) {
                    if (hasIndices)
                        refs[ks].index(ks);
                    break;
                }
                if (old[os] !== keys[ke] || old[oe] !== keys[ks])
                    break;
                refs[ks] = oldRefs[oe];
                refs[ke] = oldRefs[os];
                if (hasIndices) {
                    refs[ks].index(ks);
                    refs[ke].index(ke);
                }
                moveNodes(parentNode, oldRefs[os].range.firstNode, oldRefs[oe].range);
                if (++ks <= --ke)
                    moveNodes(parentNode, nextSibling, oldRefs[os].range);
                nextSibling = oldRefs[os].range.firstNode;
                os++;
                oe--;
            }
            // bottom up once again
            while (oe >= os && ke >= ks && old[oe] === keys[ke]) {
                if (old[oe] === keys[ke])
                    break;
                nextSibling = oldRefs[oe].range.firstNode;
                oe--, ke--;
            }
            if (ks > ke)
                for (let i = os; i <= oe; i++)
                    remover(oldRefs[i]);
            else if (os > oe) {
                // p("adding", parentNode, nextSibling)
                const range = Range;
                setRange(parentNodeRange);
                withRef(DOMRef(parentNode, nextSibling), () => {
                    for (let i = ks; i <= ke; i++)
                        createListItem(refs, keys[i], i, hasIndices, render);
                });
                setRange(range);
            }
            else {
                // inspired by ivi https://github.com/ivijs/ivi/ by Boris Kaul
                const oldMap = createMap(old);
                const winLength = ke - ks + 1, oldIndices = new Array(winLength);
                for (let i = 0; i < winLength; i++)
                    oldIndices[i] = -1; // -1 signals a new node
                let li = 0, i = 0, positionSoFar = 2147483647, matched = 0, lisIndices;
                for (i = ke; i >= ks; i--) {
                    const oi = oldMap.kind === "object" ? oldMap.map[keys[i]] : oldMap.map.get(keys[i]);
                    if (oi != null) {
                        positionSoFar = (oi < positionSoFar) ? oi : -1; // becomes -1 if nodes were re-ordered
                        oldIndices[i - ks] = oi;
                        old[oi] = null;
                        refs[i] = oldRefs[oi];
                        matched++;
                    }
                }
                if (matched !== oe - os + 1) {
                    // p("remove", {old})
                    for (let i = os; i <= oe; i++)
                        if (old[i] != null)
                            remover(oldRefs[i]);
                }
                const range = Range;
                setRange(parentNodeRange);
                withRef(DOMRef(parentNode, nextSibling), () => {
                    if (matched === 0) {
                        // p("just adding")
                        for (let i = ks; i <= ke; i++)
                            createListItem(refs, keys[i], i, hasIndices, render);
                    }
                    else {
                        // p("going LIS", JSON.stringify({pos: positionSoFar, matched, oldIndices, refs: refs.map(first), oldRefs: oldRefs.map(first), oldMap: oldMap.map}))
                        // some nodes are out of order, we need the LIS for optimal moves
                        if (positionSoFar === -1) {
                            // p("out of order")
                            // the indices of the indices of the items that are part of the
                            // longest increasing subsequence in the oldIndices list
                            lisIndices = makeLisIndices(oldIndices);
                            // p({lisIndices})
                            li = 0;
                            for (i = ks; i <= ke; i++) {
                                const oi = i - ks;
                                if (oldIndices[oi] === -1) {
                                    // p("new node")
                                    // p({bns: globalDOM.nextSibling.textContent, li})
                                    DOM.nextSibling = li < lisIndices.length ? oldRefs[oldIndices[lisIndices[li]]].range.firstNode : nextSibling;
                                    // p({ans: globalDOM.nextSibling.textContent})
                                    createListItem(refs, keys[i], i, hasIndices, render);
                                    // p(JSON.stringify({i, li, ks, oi, oldIndices}))
                                }
                                else {
                                    // p("old node")
                                    // p(JSON.stringify({i, li, ks, oi, oldIndices}))
                                    if (hasIndices)
                                        refs[i].index(i);
                                    if (li < lisIndices.length && lisIndices[li] === oi) {
                                        li++;
                                    }
                                    else {
                                        oldIndices[oi] = -1;
                                        // p({bns: globalDOM.nextSibling.textContent})
                                        DOM.nextSibling = li < lisIndices.length ? oldRefs[oldIndices[lisIndices[li]]].range.firstNode : nextSibling;
                                        // p({ans: globalDOM.nextSibling.textContent})
                                        moveNodes(DOM.parent, DOM.nextSibling, refs[i].range);
                                    }
                                }
                            }
                        }
                        else {
                            // p("in order", {ks, ke})
                            let oIi = -1;
                            for (i = ks; i <= ke; i++) {
                                const oi = i - ks;
                                os = oldIndices[oi];
                                // p({oi, os, oldIndices, map, keys, old})
                                if (os === -1) {
                                    // p("os === -1", {oIi, oi})
                                    if (oIi != null && oIi < oi) {
                                        oIi = findNext(oldIndices, oi);
                                        // p("set NS", {oi, oIi, oldIndices})
                                        DOM.nextSibling = oIi == null ? nextSibling : oldRefs[oldIndices[oIi]].range.firstNode;
                                    }
                                    createListItem(refs, keys[i], i, hasIndices, render);
                                }
                                else {
                                    if (hasIndices)
                                        refs[i].index(i);
                                }
                            }
                        }
                    }
                });
                setRange(range);
            }
            syncParents(parentNodeRange, "firstNode", oldRefs[0].range.firstNode, refs[0].range.firstNode);
            syncParents(parentNodeRange, "lastNode", oldRefs[oldRefs.length - 1].range.lastNode, refs[refs.length - 1].range.lastNode);
        }
        return { refs, keys };
    }
    function findNext(list, index) {
        for (; index < list.length; index++) {
            if (list[index] !== -1)
                return index;
        }
    }
    // Lifted from ivi https://github.com/ivijs/ivi/
    // takes a list of unique numbers (-1 is special and can
    // occur multiple times) and returns an array with the indices
    // of the items that are part of the longest increasing
    // subsequece
    // https://github.com/localvoid/ivi/blob/87794c0f45995d3368afa119560fc78825c5efb8/packages/ivi/src/vdom/reconciler.ts#L951
    function makeLisIndices(a) {
        const p = a.slice();
        // result is instantiated as an empty array to prevent instantiation with CoW backing store.
        const result = [];
        let n = 0;
        let i = 0;
        let u;
        let v;
        let j;
        result[0] = 0;
        for (; i < a.length; ++i) {
            const k = a[i];
            if (k > -1) {
                j = result[n];
                if (a[j] < k) {
                    p[i] = j;
                    result[++n] = i;
                }
                else {
                    u = 0;
                    v = n;
                    while (u < v) {
                        // eslint-disable-next-line no-bitwise
                        j = (u + v) >> 1;
                        if (a[result[j]] < k) {
                            u = j + 1;
                        }
                        else {
                            v = j;
                        }
                    }
                    if (k < a[result[u]]) {
                        if (u > 0) {
                            p[i] = result[u - 1];
                        }
                        result[u] = i;
                    }
                }
            }
        }
        v = result[n];
        while (n >= 0) {
            result[n--] = v;
            v = p[v];
        }
        if (a[0] === -1)
            result.shift();
        return result;
    }

    /*
    - cleanup, can be called from both computations and hooks, except from removing
    - removing can be called from every hook except cleanup
    - both can thus be called from reflow which can be called from render ... asap ... the ref callback.
    */
    let canCallHooks = 0;
    let asapQueue;
    function asap(cb) {
        if (canCallHooks !== 4)
            throw new Error(getErrorMessage("A001"));
        // eslint-disable-next-line @typescript-eslint/no-non-null-assertion
        asapQueue.push(cb);
    }
    function rendered(cb) {
        if (canCallHooks < 3)
            throw new Error(getErrorMessage("A001"));
        postpone("rendered", cb, Range);
    }
    function reflowed(cb) {
        if (canCallHooks < 2)
            throw new Error(getErrorMessage("A001"));
        postpone("reflowed", cb, Range);
    }
    const zoneRemoveMap = new WeakMap();
    const removeHooks = () => ({
        manager: null,
        hooks: []
    });
    function removing(cb) {
        if (Zone.removeHooks == null)
            Zone.removeHooks = removeHooks();
        const nr = Range;
        Zone.removeHooks.hooks.push(cb.length === 0
            ? result => result.push(cb())
            : result => forEachNode(nr, cb, false, result));
    }
    function Ref(cb) {
        const nr = fromParent(Range);
        zoneRemoveMap.set(nr, Zone);
        let previous = asapQueue;
        try {
            asapQueue = [];
            withRange(nr, () => {
                canCallHooks = 4;
                let res;
                try {
                    res = cb({ asap, rendered, reflowed, removing });
                }
                finally {
                    canCallHooks = 0;
                }
                emit(res);
            });
            S_.freeze(() => {
                canCallHooks = 3;
                // eslint-disable-next-line @typescript-eslint/no-non-null-assertion
                asapQueue.forEach(cb => forEachNode(nr, cb));
                canCallHooks = 0;
            });
        }
        finally {
            asapQueue = previous;
        }
    }
    function ref(cb) {
        return component(Ref, cb);
    }
    function Scheduled() {
        return { rendered: [], reflowed: [], removing: [] };
    }
    let scheduled;
    function postpone(what, fn, nr) {
        if (scheduled == null) {
            scheduled = Scheduled();
            Promise.resolve().then(processDelayedCreation);
        }
        // eslint-disable-next-line @typescript-eslint/no-explicit-any
        scheduled[what].push(fn, nr);
    }
    function processDelayedCreation() {
        // eslint-disable-next-line @typescript-eslint/no-non-null-assertion
        const { rendered, reflowed } = scheduled;
        canCallHooks = 2;
        processHookQueue(rendered);
        canCallHooks = 1;
        if (reflowed.length !== 0) {
            // todo: make sure this isn't removed by eager minifiers
            //TODO
            if (doc != null)
                doc.documentElement.clientWidth;
            processHookQueue(reflowed);
        }
        canCallHooks = 0;
        scheduled = null;
    }
    function processHookQueue(queue) {
        for (let i = 0; i < queue.length; i += 2) {
            const cb = queue[i];
            const nr = queue[i + 1];
            if (cb.length === 0) {
                const previousZone = Zone;
                setZone(zoneRemoveMap.get(nr));
                try {
                    cb();
                }
                finally {
                    setZone(previousZone);
                }
            }
            else {
                const previousRange = Range;
                const previousZone = Zone;
                setRange(nr);
                setZone(zoneRemoveMap.get(nr));
                try {
                    forEachNode(nr, cb);
                }
                finally {
                    setRange(previousRange);
                    setZone(previousZone);
                }
            }
        }
    }

    /* eslint-disable @typescript-eslint/no-explicit-any */
    const avoidAsProp = (attr) => 
    // as attr, they can take units, not as props.
    (attr === "width"
        || attr === "height"
        // This was added between Mithril v0.2 and v1, without any explanation I could find
        || attr === "href"
        // for form elements, the form property is read-only
        || attr === "form"
        // input.list is read-only
        || attr === "list"
        // so that we can remove it entirely rather than setting it to `undefined`
        || attr === "id");
    function setAttrs(el, attrs, ns, tagName) {
        if (Array.isArray(attrs))
            setAttrsArray(el, attrs, ns, tagName, true);
        else
            setAttrsObject(el, attrs, ns, tagName, false);
    }
    function setAttrsArray(el, attrs, ns, tagName, first) {
        if (first)
            dynAttrs = dynProps = null;
        attrs.forEach(a => (Array.isArray(a)
            ? setAttrsArray(el, a, ns, tagName, false)
            : setAttrsObject(el, a, ns, tagName, true)));
    }
    function setAttrsObject(el, attrs, ns, tagName, hasOverrides) {
        for (const k in attrs)
            if (hasOwn.call(attrs, k)) {
                if (k === "on")
                    setEvents(el, attrs[k]);
                else if (k === "class" || k === "className")
                    setClass(el, attrs[k]);
                else if (k === "style")
                    setStyle(el, attrs[k], hasOverrides);
                // eslint-disable-next-line @typescript-eslint/no-non-null-assertion
                else if (k === "$props")
                    Object.keys(attrs.$props).forEach(k => setAttr(el, k, attrs.$props[k], hasOverrides));
                // eslint-disable-next-line @typescript-eslint/no-non-null-assertion
                else if (k === "$attrs")
                    Object.keys(attrs.$attrs).forEach(k => setProp(el, k, attrs.$attrs[k], hasOverrides));
                else if (ns === "" && !avoidAsProp(k) && k in el)
                    setProp(el, k, attrs[k], hasOverrides);
                else
                    setAttr(el, k, attrs[k], hasOverrides);
            }
    }
    let dynProps;
    function setProp(el, k, v, hasOverrides) {
        let dyn;
        if (hasOverrides && dynProps != null && (dyn = dynProps[k]) != null) {
            S_.disposeNode(dyn);
        }
        if (typeof v === "function") {
            const { node } = S_.makeComputationNode(() => el[k] = v(), null, false, false);
            if (node != null && hasOverrides) {
                if (dynProps == null) {
                    dynProps = Object.create(null);
                }
                // eslint-disable-next-line @typescript-eslint/no-non-null-assertion
                dynProps[k] = node;
            }
        }
        else {
            el[k] = v;
        }
    }
    let dynAttrs;
    function setAttr(el, k, v, hasOverrides) {
        // TODO: handle namespaces
        // if (k.length < 6 && k.slice(0, 6) === "xlink:") el.setAttributeNS("http://www.w3.org/1999/xlink", k.slice(6), value)
        let dyn;
        if (hasOverrides && dynAttrs != null && (dyn = dynAttrs[k]) != null) {
            S_.disposeNode(dyn);
        }
        if (typeof v === "function") {
            const { node } = S_.makeComputationNode(() => {
                const value = v();
                // remove no matter what, there may be a value set by a previous attrs object
                if (value == null)
                    el.removeAttribute(k);
                else
                    el.setAttribute(k, value);
            }, null, false, false);
            if (node != null && hasOverrides) {
                if (dynAttrs == null) {
                    dynAttrs = Object.create(null);
                }
                // eslint-disable-next-line @typescript-eslint/no-non-null-assertion
                dynAttrs[k] = node;
            }
        }
        else {
            if (v == null)
                el.removeAttribute(k);
            else
                el.setAttribute(k, String(v));
        }
    }
    function setClass(el, value) {
        if (typeof value === "function") {
            S_((old) => {
                if (!skippable(old))
                    el.classList.remove(old);
                const current = value();
                if (!skippable(current))
                    el.classList.add(current);
                return current;
            }, null);
        }
        else {
            if (!skippable(value))
                el.classList.add(value);
        }
    }
    function setEvents(el, events) {
        if (typeof events === "function") {
            S_((then) => {
                const now = events();
                eventHelper(el, then, "removeEventListener");
                eventHelper(el, now, "addEventListener");
                return now;
            }, null);
        }
        else {
            eventHelper(el, events, "addEventListener");
        }
    }
    function eventHelper(el, events, method) {
        // for ... in ignores null and undefined
        for (const ev in events)
            if (hasOwn.call(events, ev)) {
                const handler = events[ev];
                if (Array.isArray(ev))
                    el[method](ev, ...handler);
                else
                    el[method](ev, handler);
            }
    }
    let dynStyle;
    let dynStyleProps;
    function setStyle(el, style, hasOverrides) {
        if (hasOverrides && dynStyle != null) {
            S_.disposeNode(dynStyle);
            dynStyle = null;
        }
        const type = typeof style;
        if (type === "object" && style != null) {
            el.style.cssText = "";
            for (const prop in style)
                if (hasOwn.call(style, prop)) {
                    if (prop.length < 2 || prop[0] === "-" && prop[1] === "-")
                        setStyleCustomProperty(el, prop, style[prop], hasOverrides);
                    else
                        setStyleProperty(el, prop, style[prop], hasOverrides);
                }
        }
        else if (type === "function") {
            if (dynStyleProps != null) {
                for (const dsp in dynStyleProps)
                    if (hasOwn.call(dynStyleProps, dsp))
                        S_.disposeNode(dynStyleProps[dsp]);
                dynStyleProps = null;
            }
            const { node } = S_.makeComputationNode(() => {
                const value = style();
                el.style.cssText = value;
            }, null, false, false);
            if (node != null && hasOverrides) {
                dynStyle = node;
            }
        }
        else {
            el.style.cssText = skippable(style) ? "" : style;
        }
    }
    function setStyleProperty(el, prop, v, hasOverrides) {
        let dyn;
        if (hasOverrides && dynStyleProps != null && (dyn = dynStyleProps[prop]) != null) {
            S_.disposeNode(dyn);
            dynStyleProps[prop] = null;
        }
        if (typeof v === "function") {
            const { node } = S_.makeComputationNode(() => {
                const value = v();
                el.style[prop] = skippable(value) ? "" : String(value);
            }, null, false, false);
            if (node != null && hasOverrides) {
                if (dynStyleProps == null)
                    dynStyleProps = {};
                dynStyleProps[prop] = node;
            }
        }
        else {
            el.style[prop] = skippable(v) ? "" : String(v);
        }
    }
    function setStyleCustomProperty(el, prop, v, hasOverrides) {
        let dyn;
        if (hasOverrides && dynStyleProps != null && (dyn = dynStyleProps[prop]) != null) {
            S_.disposeNode(dyn);
            dynStyleProps[prop] = null;
        }
        if (typeof v === "function") {
            const { node } = S_.makeComputationNode(() => {
                const value = v();
                el.style.setProperty(prop, skippable(value) ? "" : String(value));
            }, null, false, false);
            if (node != null && hasOverrides) {
                if (dynStyleProps == null)
                    dynStyleProps = {};
                dynStyleProps[prop] = node;
            }
        }
        else {
            el.style.setProperty(prop, skippable(v) ? "" : String(v));
        }
    }
    const attrsParser = /([.#])([^.#\[]+)|\[(.+?)(?:\s*=\s*("|'|)((?:\\["'\]]|.)*?)\4)?\]|((?!$))/g;
    function parseAndSetAttrs(element, s, ns) {
        attrsParser.lastIndex = 0;
        let match;
        let j = 0;
        let classes;
        while (match = attrsParser.exec(s)) {
            /* c8 ignore next */
            if (j++ === 1000) {
                console.error("attrs parser bug for " + s);
                break;
            }
            if (match[6] != null)
                throw new RangeError(`unexpected attr: ${s.slice(match.index)}`);
            if (match[1] != null) {
                if (match[1] === ".")
                    (classes || (classes = [])).push(match[2]);
                else {
                    if (ns === "")
                        element.id = match[2];
                    else
                        element.setAttribute("id", match[2]);
                }
            }
            else if (match[3] != null) {
                const key = match[3];
                const value = match[5] ? match[5].replace(/\\(["'])/g, "$1").replace(/\\\\/g, "\\") : "";
                if (key === "class")
                    (classes || (classes = [])).push(value);
                else if (avoidAsProp(key) || !(key in element))
                    element.setAttribute(key, value);
                else
                    element[key] = value;
            }
        }
        if (classes != null) {
            if (ns === "")
                element.className = classes.join(" ");
            else
                element.setAttribute("class", classes.join(" "));
        }
    }

    /* eslint-disable @typescript-eslint/no-explicit-any */
    let NS = "";
    function makeElement(selector, ns) {
        const end = selector.match(/[ \.#\[]|$/).index;
        const hasAttrs = end !== selector.length;
        const tagName = hasAttrs ? selector.slice(0, end) : selector;
        const element = (ns === "" ? doc.createElement(tagName || "div") : doc.createElementNS(ns, tagName));
        if (hasAttrs)
            parseAndSetAttrs(element, selector.slice(end), ns);
        return element;
    }
    // create the element from scratch twice before caching/cloning.
    let cacheDelay = 1;
    function getOrMakeElement(selector, ns) {
        const cached = tagCache[ns][selector] || (tagCache[ns][selector] = { delay: cacheDelay, el: null });
        if (cached.delay-- <= 0)
            return (cached.el || (cached.el = makeElement(selector, ns))).cloneNode();
        return makeElement(selector, ns);
    }
    const isPOJO = x => !skippable(x) && getProto(x) === objProto;
    // eslint-disable-next-line @typescript-eslint/ban-types
    function v$1(tagName, attrs, ...children) {
        const kind = typeof tagName;
        // eslint-disable-next-line no-constant-condition
        let candidateType;
        if (kind === "function") {
            return isPOJO(attrs)
                ? component(tagName, attrs, children)
                : component(tagName, {}, [attrs, ...children]);
        }
        else if (kind === "string") {
            let candidate = attrs;
            let l = -1;
            if (Array.isArray(candidate) && (l = candidate.length) > 0)
                candidate = candidate[0];
            return (l === 0
                || skippable(candidate)
                || (candidateType = typeof candidate) === "string"
                || candidateType === "number"
                || candidateType === "function"
                || "nodeType" in candidate)
                ? element(tagName, null, [attrs, children])
                : element(tagName, attrs, children);
        }
        else {
            throw new RangeError("string or function expected as tagName, got " + kind);
        }
    }
    // eslint-disable-next-line @typescript-eslint/ban-types
    function V(tagName, attrs, children) {
        if (typeof tagName === "function") {
            // Attrs must be a POJO, not a null-prototyped object.
            return component(tagName, isPOJO(attrs) ? attrs : {}, children);
        }
        else if (typeof tagName === "string")
            return element(tagName, attrs, children);
        else
            throw new RangeError("string or function expected as tagName, got " + typeof tagName);
    }
    function element(tagName, attrs, children) {
        const el = getOrMakeElement(tagName, NS);
        if (attrs != null)
            setAttrs(el, attrs, NS, tagName);
        withRef(DOMRef(el, null), () => {
            withoutRange(() => {
                emit(children);
            });
        });
        return el;
    }

    // lib/* is generated from src/* at compile-time

    var all = /*#__PURE__*/Object.freeze({
        __proto__: null,
        keyed: keyed,
        ref: ref,
        boot: boot,
        S: S_,
        V: V,
        v: v$1
    });

    const {v, S} = all;
    Object.assign(v, all);

    exports.S = S;
    exports.v = v;

    Object.defineProperty(exports, '__esModule', { value: true });

}));
