/* eslint-disable @typescript-eslint/no-explicit-any */
/* eslint-disable @typescript-eslint/no-non-null-assertion */
import { parseAndSetAttrs, setAttrs } from "./attrs.js";
import { doc, tagCache } from "./env.js";
import { DOMRef, component, emit, withRef, withoutRange } from "./render.js";
import { getProto, objProto, skippable } from "./util.js";
export { SVG, V };
export { svg, v };
export { cacheDelay, setCacheDelay };
let NS = "";
function makeElement(selector, ns) {
    const end = selector.match(/[ \.#\[]|$/).index;
    const hasAttrs = end !== selector.length;
    const tagName = hasAttrs ? selector.slice(0, end) : selector;
    const element = (ns === "" ? doc.createElement(tagName || "div") : doc.createElementNS(ns, tagName));
    if (hasAttrs)
        parseAndSetAttrs(element, selector.slice(end), ns);
    return element;
}
// create the element from scratch twice before caching/cloning.
let cacheDelay = 1;
function setCacheDelay(n) {
    cacheDelay = n;
}
function getOrMakeElement(selector, ns) {
    const cached = tagCache[ns][selector] || (tagCache[ns][selector] = { delay: cacheDelay, el: null });
    if (cached.delay-- <= 0)
        return (cached.el || (cached.el = makeElement(selector, ns))).cloneNode();
    return makeElement(selector, ns);
}
const isPOJO = x => !skippable(x) && getProto(x) === objProto;
function svg(selector, ...args) {
    NS = "http://www.w3.org/2000/svg";
    try {
        return v(selector, ...(args));
    }
    finally {
        NS = "";
    }
}
// eslint-disable-next-line @typescript-eslint/ban-types
function v(tagName, attrs, ...children) {
    const kind = typeof tagName;
    // eslint-disable-next-line no-constant-condition
    let candidateType;
    if (kind === "function") {
        return isPOJO(attrs)
            ? component(tagName, attrs, children)
            : component(tagName, {}, [attrs, ...children]);
    }
    else if (kind === "string") {
        let candidate = attrs;
        let l = -1;
        if (Array.isArray(candidate) && (l = candidate.length) > 0)
            candidate = candidate[0];
        return (l === 0
            || skippable(candidate)
            || (candidateType = typeof candidate) === "string"
            || candidateType === "number"
            || candidateType === "function"
            || "nodeType" in candidate)
            ? element(tagName, null, [attrs, children])
            : element(tagName, attrs, children);
    }
    else {
        throw new RangeError("string or function expected as tagName, got " + kind);
    }
}
function SVG(selector, ...args) {
    NS = "http://www.w3.org/2000/svg";
    try {
        return V(selector, ...args);
    }
    finally {
        NS = "";
    }
}
// eslint-disable-next-line @typescript-eslint/ban-types
function V(tagName, attrs, children) {
    if (typeof tagName === "function") {
        // Attrs must be a POJO, not a null-prototyped object.
        return component(tagName, isPOJO(attrs) ? attrs : {}, children);
    }
    else if (typeof tagName === "string")
        return element(tagName, attrs, children);
    else
        throw new RangeError("string or function expected as tagName, got " + typeof tagName);
}
function element(tagName, attrs, children) {
    const el = getOrMakeElement(tagName, NS);
    if (attrs != null)
        setAttrs(el, attrs, NS, tagName);
    withRef(DOMRef(el, null), () => {
        withoutRange(() => {
            emit(children);
        });
    });
    return el;
}
//# sourceMappingURL=v.js.map